#include<stdio.h>
#include<stdlib.h>
#include<math.h>
#include<string.h>

#define pi 3.141592653589793
#define bufsize 1000
#define Nmax 1000000

struct vertex {
  int read_from_file;
  double x,y,z;
};

struct face {
  int nvertices,n_v_first,dir,cluster,next,read_from_file;
  double nx,ny,nz;
};

struct edge {
  int f1,f2,v1,v2,d1,d2;
  double x1,y1,z1,x2,y2,z2,gamma[6];
};

struct triangle {
  int f;
  double x1,y1,z1,x2,y2,z2,x3,y3,z3,n[6];
};

struct vertex* vertices;
struct face* faces;
struct edge* edges;
struct triangle* triangles;
int* n_v;
int Nvertices,Nfaces,Nedges,Ntriangles,n_v_max;
double eps0,Poisson;
char *filename_parameters, *filename_input, *filename_output;

void get_eps3d(double x, double y, double z, int Nedges, int Ntriangles, 
 struct edge* edges, struct triangle* triangles, double eps0, double Poisson, 
 double* result_eps, double* result_chi);
void set_edges_and_triangles();

void read_parameters(char* filename);
void calc_strain(char* filename_in, char* filename_out);

int main(int argc, char* argv[]) {
  if(argc!=4) fprintf(stderr,"Error!\n  main: wrong number of arguments.\n"),exit(1);
  filename_parameters=argv[1];
  filename_input=argv[2];
  filename_output=argv[3];
  read_parameters(filename_parameters);
  set_edges_and_triangles();
  calc_strain(filename_input,filename_output);
  free(vertices);
  free(faces);
  free(edges);
  free(triangles);
  free(n_v);
  return 0;
}


// -----------------------

double get_phi_rod(double x, double y, double z, double x1, double y1, double z1, 
 double x2, double y2, double z2);
double get_solid_angle(double x, double y, double z, double x1, double y1, double z1,
 double x2, double y2, double z2, double x3, double y3, double z3);
void set_gamma_for_all_edges();
void set_n_for_all_triangles();
double get_cluster_volume(int f);
void get_face_area(int f, double* nx, double* ny, double* nz);

struct edge_temp {int v1,v2,f,d;};
int sort_function_for_edges_temp(const void*,const void*);


void get_eps3d(double x, double y, double z, int Nedges, int Ntriangles, 
 struct edge* edges, struct triangle* triangles, double eps0, double Poisson, 
 double* result_eps, double* result_chi) {
  int i,k;
  double lambda,epsxx,epsyy,epszz,epsxy,epsxz,epsyz,omega,phi,chi;
  lambda=eps0*(1+Poisson)/(4*pi*(1-Poisson));
  epsxx=epsyy=epszz=epsxy=epsxz=epsyz=chi=0;
  for (i=1; i<=Ntriangles; i++) {
    omega=get_solid_angle(x,y,z,triangles[i].x1,triangles[i].y1,triangles[i].z1,
          triangles[i].x2,triangles[i].y2,triangles[i].z2,triangles[i].x3,triangles[i].y3,triangles[i].z3);
	epsxx-=lambda*triangles[i].n[0]*omega;
	epsyy-=lambda*triangles[i].n[1]*omega;
	epszz-=lambda*triangles[i].n[2]*omega;
	epsxy-=lambda*triangles[i].n[3]*omega;
	epsxz-=lambda*triangles[i].n[4]*omega;
	epsyz-=lambda*triangles[i].n[5]*omega;
	chi-=omega/(4*pi);
  }
  if (fabs(chi)<1e-8) chi=0;
  if (fabs(chi-1)<1e-8) chi=1;
  epsxx-=eps0*chi;
  epsyy-=eps0*chi;
  epszz-=eps0*chi;
  for (k=1; k<=Nedges; k++) {
    phi=get_phi_rod(x,y,z,edges[k].x1,edges[k].y1,edges[k].z1,edges[k].x2,edges[k].y2,edges[k].z2);
	epsxx-=lambda*edges[k].gamma[0]*phi;
	epsyy-=lambda*edges[k].gamma[1]*phi;
	epszz-=lambda*edges[k].gamma[2]*phi;
	epsxy-=lambda*edges[k].gamma[3]*phi;
	epsxz-=lambda*edges[k].gamma[4]*phi;
	epsyz-=lambda*edges[k].gamma[5]*phi;
  }
  result_eps[0]=epsxx;
  result_eps[1]=epsyy;
  result_eps[2]=epszz;
  result_eps[3]=epsxy;
  result_eps[4]=epsxz;
  result_eps[5]=epsyz;
  *result_chi=chi;
}


double get_phi_rod(double x, double y, double z, double x1, double y1, double z1, 
 double x2, double y2, double z2) {
  double r1,r2,L;
  L=sqrt((x1-x2)*(x1-x2)+(y1-y2)*(y1-y2)+(z1-z2)*(z1-z2));
  if (L==0) return 0;
  r1=sqrt((x-x1)*(x-x1)+(y-y1)*(y-y1)+(z-z1)*(z-z1));
  r2=sqrt((x-x2)*(x-x2)+(y-y2)*(y-y2)+(z-z2)*(z-z2));
  if (r1+r2-L<=0) return 1000;
  else return log((r1+r2+L)/(r1+r2-L));
}

double get_solid_angle(double x, double y, double z, double x1, double y1, double z1,
 double x2, double y2, double z2, double x3, double y3, double z3) {
  double ax,ay,az,bx,by,bz,cx,cy,cz,a,b,c,ab,ac,bc,P,Q;
  ax=x1-x; ay=y1-y; az=z1-z;
  bx=x2-x; by=y2-y; bz=z2-z;
  cx=x3-x; cy=y3-y; cz=z3-z;
  a=sqrt(ax*ax+ay*ay+az*az);
  b=sqrt(bx*bx+by*by+bz*bz);
  c=sqrt(cx*cx+cy*cy+cz*cz);
  ab=ax*bx+ay*by+az*bz;
  ac=ax*cx+ay*cy+az*cz;
  bc=bx*cx+by*cy+bz*cz;
  P=ax*by*cz+ay*bz*cx+az*bx*cy-ax*bz*cy-ay*bx*cz-az*by*cx;  // [axb]*c
  Q=a*b*c+ab*c+ac*b+bc*a;
  if (P==0 && Q==0) return 0;
  else return 2*atan2(P,Q);
}


void set_edges_and_triangles() {
  int i,k,v,f,et,vt,nv1,nv2,v1,v2,d,f1,f2,fx,tempc,temp,Nedges_temp;
  double nx,ny,nz,lx,ly,lz,l,x1,y1,z1;
  struct edge_temp *edges_temp;
  // At this point, it is assumed that arrays "vertices", "faces", "n_v" are created; variables "Nvertices", "Nfaces", "n_v_max" are set;
  // arrays "vertices" and "n_v" are filled; and in the array "faces" fields "nvertices" and "n_v_first" are filled.
  // Now check arrays "vertices", "faces", "n_v".
  if (Nfaces<4) fprintf(stderr,"Error!\n  set_edges_and_triangles: number of faces is less than 4.\n"),exit(1);
  if (Nvertices<4) fprintf(stderr,"Error!\n  set_edges_and_triangles: number of vertices is less than 4.\n"),exit(1);
  for (f=1; f<=Nfaces; f++)
    if (faces[f].nvertices<3) 
	  fprintf(stderr,"Error!\n  set_edges_and_triangles: there are less than three vertices in face %d.\n",f),exit(1);
  for (vt=0; vt<n_v_max; vt++) 
    if (n_v[vt]<1 || n_v[vt]>Nvertices)
	  fprintf(stderr,"Error!\n  set_edges_and_triangles: internal error.\n"),exit(1);
  for (f=1; f<=Nfaces; f++) {
    nv1=faces[f].n_v_first;
	nv2=faces[f].n_v_first+faces[f].nvertices-1;
	if (nv1<0 || nv2<0 || nv1>=n_v_max || nv2>=n_v_max) fprintf(stderr,"Error!\n  set_edges_and_triangles: internal error.\n"),exit(1);
  }
  // Make an array of edges "edges_temp".
  Nedges_temp=0;
  for (f=1; f<=Nfaces; f++)
    Nedges_temp+=faces[f].nvertices;
  if (Nedges_temp>Nmax*2) fprintf(stderr,"Error!\n  set_edges_and_triangles: there are too many edges.\n"),exit(1);
  edges_temp=(struct edge_temp*) malloc(Nedges_temp*sizeof(struct edge_temp));
  if (edges_temp==NULL) fprintf(stderr,"Error!\n  set_edges_and_triangles: can't allocate memory.\n"),exit(1);
  et=0;
  for (f=1; f<=Nfaces; f++)
    for (vt=0; vt<faces[f].nvertices; vt++) {
	  nv1=faces[f].n_v_first+vt;
	  nv2=faces[f].n_v_first+(vt+1)%(faces[f].nvertices);
	  if (nv1<0 || nv2<0 || nv1>=n_v_max || nv2>=n_v_max) fprintf(stderr,"Error!\n  set_edges_and_triangles: internal error.\n"),exit(1);
	  v1=n_v[nv1];
	  v2=n_v[nv2];
	  if (v1==v2) fprintf(stderr,"Error!\n  set_edges_and_triangles: there are two equal adjacent vertices in face %d.\n",f),exit(1);
	  if (vertices[v1].x==vertices[v2].x && vertices[v1].y==vertices[v2].y && vertices[v1].z==vertices[v2].z) 
	   fprintf(stderr,"Error!\n  set_edges_and_triangles: there are two equal adjacent vertices in face %d.\n",f),exit(1);
	  if (v1<v2) d=1;
	  else {d=-1; temp=v1; v1=v2; v2=temp;}
	  edges_temp[et].v1=v1;
	  edges_temp[et].v2=v2;
	  edges_temp[et].f=f;
	  edges_temp[et].d=d;
	  et++;
	}
  if (et!=Nedges_temp) fprintf(stderr,"Error!\n  set_edges_and_triangles: internal error.\n"),exit(1);
  // Sort the array "edges_temp"; check that each edge enters twice.
  qsort((void*)edges_temp,Nedges_temp,sizeof(struct edge_temp),sort_function_for_edges_temp);
  for (et=0; et<Nedges_temp-1; et++) {
    if (et%2==0 && (edges_temp[et].v1!=edges_temp[et+1].v1 || edges_temp[et].v2!=edges_temp[et+1].v2))
	  fprintf(stderr,"Error!\n  set_edges_and_triangles: there is an edge (%d,%d) belonging to only one face.\n",
	   edges_temp[et].v1,edges_temp[et].v2),exit(1);
    if (et%2==1 && edges_temp[et].v1==edges_temp[et+1].v1 && edges_temp[et].v2==edges_temp[et+1].v2)
	  fprintf(stderr,"Error!\n  set_edges_and_triangles: three or more faces share an edge (%d,%d).\n",
	   edges_temp[et].v1,edges_temp[et].v2),exit(1);
  }
  et=Nedges_temp-1;
  if (et%2==0) 
	fprintf(stderr,"Error!\n  set_edges_and_triangles: there is an edge (%d,%d) belonging to only one face.\n",
	 edges_temp[et].v1,edges_temp[et].v2),exit(1);
  // Make the array "edges"; fill all fields in it, except "gamma".
  Nedges=Nedges_temp/2;
  if (Nedges>Nmax) fprintf(stderr,"Error!\n  set_edges_and_triangles: there are too many edges.\n"),exit(1);
  edges=(struct edge*) malloc((Nedges+1)*sizeof(struct edge));
  if (edges==NULL) fprintf(stderr,"Error!\n  set_edges_and_triangles: can't allocate memory.\n"),exit(1);
  for (k=1; k<=Nedges; k++) {
    et=(k-1)*2;
	v1=edges_temp[et].v1;
	v2=edges_temp[et].v2;
	edges[k].f1=edges_temp[et].f;
	edges[k].f2=edges_temp[et+1].f;
	edges[k].d1=edges_temp[et].d;
	edges[k].d2=edges_temp[et+1].d;
	edges[k].v1=v1;
	edges[k].v2=v2;
	edges[k].x1=vertices[v1].x;
	edges[k].x2=vertices[v2].x;
	edges[k].y1=vertices[v1].y;
	edges[k].y2=vertices[v2].y;
	edges[k].z1=vertices[v1].z;
	edges[k].z2=vertices[v2].z;
  }
  free(edges_temp);
  // Set initial values for fields "dir", "cluster", "next" of the array "faces".
  for (f=1; f<=Nfaces; f++) {
    faces[f].dir=1;
	faces[f].cluster=f;
	faces[f].next=f;
  }
  // Form clusters (update fields "dir", "cluster", "next" of the array "faces").
  for (k=1; k<=Nedges; k++) {
    f1=edges[k].f1; f2=edges[k].f2;
	if (f1==f2) fprintf(stderr,"Error!\n  set_edges_and_triangles: incorrect configuration of face %d.\n",f1),exit(1);
	if (faces[f1].cluster==faces[f2].cluster && faces[f1].dir*faces[f2].dir*edges[k].d1*edges[k].d2!=-1) 
	 fprintf(stderr,"Error!\n  set_edges_and_triangles: incorrect configuration of a fragment including faces %d,%d.\n",f1,f2),exit(1);
    if (faces[f1].cluster!=faces[f2].cluster) {
	  d=-faces[f1].dir*faces[f2].dir*edges[k].d1*edges[k].d2;
	  fx=f2;
	  for (tempc=0; tempc<=Nfaces; tempc++) {
        if (tempc==Nfaces) fprintf(stderr,"Error!\n  set_edges_and_triangles: internal error.\n"),exit(1);
		faces[fx].cluster=faces[f1].cluster;
		faces[fx].dir*=d;
		if (faces[fx].next==f2) break;
		fx=faces[fx].next;
	  }
	  faces[fx].next=faces[f1].next;
	  faces[f1].next=f2;
	}
  }
  // For each cluster: calculate its volume; if the volume is negative, correct the field "dir" of the array "faces".
  for (f=1; f<=Nfaces; f++)
    if (faces[f].cluster==f)
	  if (get_cluster_volume(f)<0) {
	    fx=f;
	    for (tempc=0; tempc<=Nfaces; tempc++) {
          if (tempc==Nfaces) fprintf(stderr,"Error!\n  set_edges_and_triangles: internal error.\n"),exit(1);
		  faces[fx].dir*= -1;
		  fx=faces[fx].next;
		  if (fx==f) break;
	    }
	  }
  // Calculate fields "nx", "ny", "nz" (a normal unit vector) of the array "faces".
  for (f=1; f<=Nfaces; f++) {
    get_face_area(f,&nx,&ny,&nz);
	l=sqrt(nx*nx+ny*ny+nz*nz);
	if (l==0) fprintf(stderr,"Error!\n  set_edges_and_triangles: strange face %d - its area is zero.\n",f),exit(1);
	nx=nx/l; ny=ny/l; nz=nz/l;
	faces[f].nx=nx;
	faces[f].ny=ny;
	faces[f].nz=nz;
  }
  for (k=1; k<=Nedges; k++) {
    lx=edges[k].x1-edges[k].x2;
    ly=edges[k].y1-edges[k].y2;
    lz=edges[k].z1-edges[k].z2;
    l=sqrt(lx*lx+ly*ly+lz*lz);
    f=edges[k].f1;
    nx=faces[f].nx; ny=faces[f].ny; nz=faces[f].nz;
    if (fabs(lx*nx+ly*ny+lz*nz)>1e-8*l) fprintf(stderr,"Error!\n  face %d if not flat.\n",f),exit(1);
    f=edges[k].f2;
    nx=faces[f].nx; ny=faces[f].ny; nz=faces[f].nz;
    if (fabs(lx*nx+ly*ny+lz*nz)>1e-8*l) fprintf(stderr,"Error!\n  face %d if not flat.\n",f),exit(1);
  }
  // Make the array "triangles"; fill all its fields, except "n".
  Ntriangles=0;
  for (f=1; f<=Nfaces; f++)
    Ntriangles+=faces[f].nvertices-2;
  if (Ntriangles>Nmax) fprintf(stderr,"Error!\n  set_edges_and_triangles: there are too many triangles.\n"),exit(1);
  triangles=(struct triangle*) malloc((Ntriangles+1)*sizeof(struct triangle));
  if (triangles==NULL) fprintf(stderr,"Error!\n  set_edges_and_triangles: can't allocate memory.\n"),exit(1);
  i=1;
  for (f=1; f<=Nfaces; f++) {
    if (abs(faces[f].dir)!=1) fprintf(stderr,"Error!\n  set_edges_and_triangles: internal error.\n"),exit(1);
    nv1=faces[f].n_v_first;
    nv2=faces[f].n_v_first+faces[f].nvertices-1;
    if (nv1<0 || nv2<0 || nv1>=n_v_max || nv2>=n_v_max) fprintf(stderr,"Error!\n  set_edges_and_triangles: internal error.\n"),exit(1);
    v=n_v[nv1];
    x1=vertices[v].x; y1=vertices[v].y; z1=vertices[v].z;
    for (vt=nv1+1; vt<nv2; vt++) {
      if (i>Ntriangles || i<1) fprintf(stderr,"Error!\n  set_edges_and_triangles: internal error.\n"),exit(1);
      triangles[i].x1=x1;
      triangles[i].y1=y1;
      triangles[i].z1=z1;
      if (faces[f].dir>0) v=n_v[vt+1]; else v=n_v[vt];
      triangles[i].x2=vertices[v].x;
      triangles[i].y2=vertices[v].y;
      triangles[i].z2=vertices[v].z;
      if (faces[f].dir>0) v=n_v[vt]; else v=n_v[vt+1];
      triangles[i].x3=vertices[v].x;
      triangles[i].y3=vertices[v].y;
      triangles[i].z3=vertices[v].z;
      triangles[i].f=f;
      i++;
    }
  }
  if (i!=Ntriangles+1) fprintf(stderr,"Error!\n  set_edges_and_triangles: internal error.\n"),exit(1);
  // Make sure that (f1,v1,f2,v2) go in clockwise direction (looking from outside the cluster)
  for (k=1; k<=Nedges; k++)
   if (edges[k].d1*faces[edges[k].f1].dir==1) {
    f=edges[k].f1;
    edges[k].f1=edges[k].f2;
    edges[k].f2=f;
   }
  // Calculate the field "gamma" of the array "edges".
  set_gamma_for_all_edges();
  // Calculate the field "n" of the array "triangles".
  set_n_for_all_triangles();
}

double get_cluster_volume(int f) {
  int fx,tempc,nv1,nv2,vt,v;
  double V,Vf,x1,x2,x3,y1,y2,y3,z1,z2,z3;
  V=0; fx=f;
  for (tempc=0; tempc<=Nfaces; tempc++) {
    if (tempc==Nfaces) fprintf(stderr,"Error!\n  get_cluster_volume: internal error.\n"),exit(1);
	Vf=0;
    nv1=faces[fx].n_v_first;
	nv2=faces[fx].n_v_first+faces[fx].nvertices-1;
	if (nv1<0 || nv2<0 || nv1>=n_v_max || nv2>=n_v_max) fprintf(stderr,"Error!\n  get_cluster_volume: internal error.\n"),exit(1);
	v=n_v[nv1];
	x1=vertices[v].x; y1=vertices[v].y; z1=vertices[v].z;
	for (vt=nv1+1; vt<nv2; vt++) {
	  v=n_v[vt];
	  x2=vertices[v].x; y2=vertices[v].y; z2=vertices[v].z;
	  v=n_v[vt+1];
	  x3=vertices[v].x; y3=vertices[v].y; z3=vertices[v].z;
	  Vf+=(z1+z2+z3)*((x2-x1)*(y3-y2)-(x3-x2)*(y2-y1))/6;
	}
	V+=Vf*faces[fx].dir;
	fx=faces[fx].next;
	if (fx==f) break;
  }
  return V;
}

void get_face_area(int f, double* nx, double* ny, double* nz) {
  int nv1,nv2,vt,v;
  double Sx,Sy,Sz,x1,x2,x3,y1,y2,y3,z1,z2,z3;
  Sx=Sy=Sz=0;
  nv1=faces[f].n_v_first;
  nv2=faces[f].n_v_first+faces[f].nvertices-1;
  if (nv1<0 || nv2<0 || nv1>=n_v_max || nv2>=n_v_max) fprintf(stderr,"Error!\n  get_face_area: internal error.\n"),exit(1);
  v=n_v[nv1];
  x1=vertices[v].x; y1=vertices[v].y; z1=vertices[v].z;
  for (vt=nv1+1; vt<nv2; vt++) {
	v=n_v[vt];
	x2=vertices[v].x; y2=vertices[v].y; z2=vertices[v].z;
	v=n_v[vt+1];
	x3=vertices[v].x; y3=vertices[v].y; z3=vertices[v].z;
	Sx+=((y2-y1)*(z3-z2)-(y3-y2)*(z2-z1))/2;
	Sy+=((z2-z1)*(x3-x2)-(z3-z2)*(x2-x1))/2;
	Sz+=((x2-x1)*(y3-y2)-(x3-x2)*(y2-y1))/2;
  }
  *nx=Sx*faces[f].dir;
  *ny=Sy*faces[f].dir;
  *nz=Sz*faces[f].dir;
}


void set_gamma_for_all_edges() {
  int k,f1,f2;
  double nx1,ny1,nz1,nx2,ny2,nz2,lx,ly,lz,l,bx1,by1,bz1,bx2,by2,bz2;
  for (k=1; k<=Nedges; k++) {
    f1=edges[k].f1; f2=edges[k].f2;
    if (f1<1 || f1>Nfaces || f2<1 || f2>Nfaces) fprintf(stderr,"Error!\n  set_gamma_for_all_edges: internal error.\n"),exit(1);
    nx1=faces[f1].nx;
    ny1=faces[f1].ny;
    nz1=faces[f1].nz;
    nx2=faces[f2].nx;
    ny2=faces[f2].ny;
    nz2=faces[f2].nz;
    lx=edges[k].x1-edges[k].x2;
    ly=edges[k].y1-edges[k].y2;
    lz=edges[k].z1-edges[k].z2;
    l=sqrt(lx*lx+ly*ly+lz*lz);
    if (l==0) fprintf(stderr,"Error!\n  set_gamma_for_all_edges: internal error.\n"),exit(1);
    lx=lx/l; ly=ly/l; lz=lz/l;
    bx1= ly*nz1-lz*ny1;
    by1= lz*nx1-lx*nz1;
    bz1= lx*ny1-ly*nx1;
    bx2=-ly*nz2+lz*ny2;
    by2=-lz*nx2+lx*nz2;
    bz2=-lx*ny2+ly*nx2;
    edges[k].gamma[0]= nx1*bx1+nx2*bx2;
    edges[k].gamma[1]= ny1*by1+ny2*by2;
    edges[k].gamma[2]= nz1*bz1+nz2*bz2;
    edges[k].gamma[3]= nx1*by1+nx2*by2;
    edges[k].gamma[4]= nx1*bz1+nx2*bz2;
    edges[k].gamma[5]= ny1*bz1+ny2*bz2;
  }
}

void set_n_for_all_triangles() {
  int i,f;
  double nx,ny,nz;
  for (i=1; i<=Ntriangles; i++) {
    f=triangles[i].f;
	if (f<1 || f>Nfaces) fprintf(stderr,"Error!\n  set_n_for_all_triangles: internal error.\n"),exit(1);
	nx=faces[f].nx;
	ny=faces[f].ny;
	nz=faces[f].nz;
	triangles[i].n[0]=nx*nx;
	triangles[i].n[1]=ny*ny;
	triangles[i].n[2]=nz*nz;
	triangles[i].n[3]=nx*ny;
	triangles[i].n[4]=nx*nz;
	triangles[i].n[5]=ny*nz;
  }
}

int sort_function_for_edges_temp(const void* a, const void* b) {
  int v1a,v2a,v1b,v2b;
  v1a=(*(struct edge_temp*)a).v1;
  v2a=(*(struct edge_temp*)a).v2;
  v1b=(*(struct edge_temp*)b).v1;
  v2b=(*(struct edge_temp*)b).v2;
  if (v1a<v1b) return -1;
  if (v1a>v1b) return 1;
  if (v2a<v2b) return -1;
  if (v2a>v2b) return 1;
  return 0;
}

// ------------------------


int alex_fgets (char* s, int n, FILE* in);

void calc_strain(char* filename_in, char* filename_out) {
  double eps[6],chi,x,y,z;
  FILE *in, *out;
  char buf[bufsize];
  int f,n;
  in=fopen(filename_in,"r");
  if (in==NULL) fprintf(stderr,"Error!\n  calc_strain: can't open the file %s for reading.\n",filename_in),exit(1);
  out=fopen(filename_out,"w");
  if (out==NULL) fprintf(stderr,"Error!\n  calc_strain: can't open the file %s for writing.\n",filename_out),exit(1);
  while (1) {
    f=alex_fgets(buf,bufsize,in);
	if (f==0) break;
	n=sscanf(buf,"%lf%lf%lf",&x,&y,&z);
	if (n!=3) fprintf(out,"%s\n",buf);
	else {
	  get_eps3d(x,y,z,Nedges,Ntriangles,edges,triangles,eps0,Poisson,eps,&chi);
      fprintf(out,"%s   %.2lf  %10lf  %10lf  %10lf  %10lf  %10lf  %10lf\n",
	          buf,chi,eps[0],eps[1],eps[2],eps[3],eps[4],eps[5]);
	}
  }
  fclose(in);
  fclose(out);
}

int alex_fgets (char* s, int n, FILE* in) {
  int i,f;
  char c;
  if (n<2) return 0;
  i=0; f=0;
  while (1) {
    if (i>=n-1) break;
	c=fgetc(in);
	if (c==EOF) break;
	if (c=='\r') continue;
	f=1;
	if (c=='\n') break;
	s[i]=c;
	i++;
  }
  s[i]=0;
  return f;
}


// ------------------------

int vertices_done, faces_done, eps0_done, Poisson_done;
int n_v_Max, next_word_already_read;
char s[bufsize+1], format_string[30], first_symbol;
FILE* in;

void read_vertices();
void read_one_vertex();
void read_faces();
int  start_new_face();
void read_vertex_for_face(int f);

void read_next_word();
void read_next_word_1();
double read_double_from_file();
int read_integer_from_file();
int read_integer_after_at();

// issue an error message "msg" and stop the program.
#define rp_error(msg) fprintf(stderr,"Error!\n  read_parameters: " msg ".\n"),exit(1)
// the same with an argument "p1" of the error message
#define rp_error1(msg,p1) fprintf(stderr,"Error!\n  read_parameters: " msg ".\n",p1),exit(1)
// the same with two arguments "p1", "p2"
#define rp_error2(msg,p1,p2) fprintf(stderr,"Error!\n  read_parameters: " msg ".\n",p1,p2),exit(1)

//  check whether the last read word coincides with the string "x"
#define word_match(x) (!strcmp(s,x))


// This function reads the file "filename" and picks all information about the structure from this file.
void read_parameters(char* filename) {
  int v,f,vt,v1,v2;
  // create the format string for reading of words (taking into account that the buffer can store no more than (bufsize-1) characters)
  sprintf(format_string,"%%%ds",bufsize-1);
  // open the file "filename" using the handler "in"
  in=fopen(filename,"r");
  if (in==NULL) rp_error1("can't open the file %s for reading",filename);
  // tell the program that sections "_vertices", "_faces", "_eps0", "_Poisson"  haven't read yet, 
  vertices_done=faces_done=eps0_done=Poisson_done=0;
  // and that the next word has not read yet
  next_word_already_read=0;
  // start reading the file, word by word
  while (1) {
    read_next_word();
    // here we pick up only keywords (i. e. words that begin from '_'); the other important words will be read by other functions
    // if there is a keyword "_end", then terminate reading of the file
    if (word_match("_end"))      break;
    // for other known keywords, do appropriate things and return to the beginning of the loop
    if (word_match("_vertices")) {read_vertices(); continue;}
    if (word_match("_faces"))    {read_faces();    continue;}
    if (word_match("_eps0"))     {eps0   =read_double_from_file();    eps0_done=1; continue;}
    if (word_match("_Poisson"))  {Poisson=read_double_from_file(); Poisson_done=1; continue;}
    // if the word is an unknown keyword, it will be ignored, but a warning will be issued
    if (first_symbol=='_') fprintf(stderr,"Warning!\n  unknown keyword '%s' in input file '%s'.\n",s,filename);
    // if here appears a word that is not a keyword, it wil be ignored
  }
  // after keyword "_end", close the file
  fclose(in);
  // check whether all necessary sections have read from the file
  if (!vertices_done) rp_error("keyword '_vertices' not found");
  if (!faces_done)    rp_error("keyword '_faces' not found");
  if (!eps0_done)     rp_error("keyword '_eps0' not found");
  if (!Poisson_done)  rp_error("keyword '_Poisson' not found");
  // check that each face has at least 3 vertices,
  for (f=1; f<=Nfaces; f++)
    if (faces[f].nvertices<3) 
	     rp_error1("there are less than three vertices in face %d",f);
  // and that all numbers of vertices are correct
  for (vt=0; vt<n_v_max; vt++) 
    if (n_v[vt]<1 || n_v[vt]>Nvertices)
	     rp_error1("wrong number of vertice '%d'",n_v[vt]);
  // output for debugging
  /*
  printf("eps0=%lf   Poisson=%lf   Nvertices=%d,   Nfaces=%d,   n_v_max=%d.\n",eps0,Poisson,Nvertices,Nfaces,n_v_max);
  for (v=1; v<=Nvertices; v++)
    printf("vertex %d:  x=%lf  y=%lf  z=%lf\n",v,vertices[v].x,vertices[v].y,vertices[v].z);
  for (f=1; f<=Nfaces; f++) {
    printf("face %d: ",f);
	   v1=faces[f].n_v_first; v2=v1+faces[f].nvertices;
	   for (vt=v1; vt<v2; vt++) printf(" %d",n_v[vt]);
	   printf("\n");
  }
  */
  // At this point, arrays "vertices", "faces", "n_v" are created; variables "Nvertices", "Nfaces", "n_v_max", "eps0", "Poisson" are set;
  // arrays "vertices" and "n_v" are filled; and in the array "faces" fields "nvertices" and "n_v_first" are filled.
}


// This function is called when the keyword "_vertices" is read from the file.
// The function reads the whole section "_vertices" (i. e. a part of file between keyword "_vertices" and the next keyword),
// extract information from this section, and put this information into array "vertices" and variable "Nvertices".
void read_vertices() {
  int v;
  // check whether section "_vertices" was appeared earlier in the file (that is prohibited -- this section should appear only once)
  if (vertices_done) rp_error("section '_vertices' appears twice");
  // read the number of vertices
  Nvertices=read_integer_from_file();
  // check whether this number is correct,
	 if (Nvertices<4 || Nvertices>Nmax) rp_error("number of vertices is out of range");
  // create the array "vertices"
  vertices=(struct vertex*) malloc((Nvertices+1)*sizeof(struct vertex));
  if (vertices==NULL) rp_error("can't allocate memory");
  // tell the program that no one of vertices is read yet
  for (v=1; v<=Nvertices; v++) vertices[v].read_from_file=0;
  // now read all the rest words of section "_vertices"
  while (1) {
    read_next_word();
    // if the word begins with '_', then section "_vertices" is over (this word is a keyword that starts the next section)
    if (first_symbol=='_') break;
    // if the word begins with '@', then information about a new vertex is started here; let's read this information
    if (first_symbol=='@') read_one_vertex();
  }
  // after reading the whole section "_verties", let's check that all vertices were appeared in this section
  for (v=1; v<=Nvertices; v++)
		 if (vertices[v].read_from_file==0)
    rp_error1("vertice %d not found",v);
  // in conclusion, we let the program know that the section "_faces" is already read from file,
  vertices_done=1;
  // and that the first word of the next section is read from file
  next_word_already_read=1;
}

// This function is called when a word is read (in section "_vertices") that begins with symbol '@'.
// It means that information about a new vertex starts here, and the number of this vertex is written after '@'.
// The function picks this number of vertex, then reads its (x,y,z)-coordinates from the file, and saves them in array "vertices".
void read_one_vertex() {
  int v;
  // pick the number of the vertex (variable "v")
  v=read_integer_after_at();
  // check whether this number is correct,
		if (v<1 || v>Nvertices) rp_error1("wrong number of vertice '%d'",v);
  // and whether this face was already appear in the file (that is prohibited -- each face should appear only once)
		if (vertices[v].read_from_file!=0) rp_error1("vertice %d appears twice",v);
  //read three coordinates (of vertex v) from the file, and save them in the array "vertices"
  vertices[v].x=read_double_from_file();
  vertices[v].y=read_double_from_file();
  vertices[v].z=read_double_from_file();
  // finally, tell the program that this vertex is already read from the file
		vertices[v].read_from_file=1;
}

// This function is called when the keyword "_faces" is read from the file.
// The function reads the whole section "_faces" (i. e. a part of file between keyword "_faces" and the next keyword),
// extract information from this section, and put this information into arrays "faces", "n_v", and variables "Nfaces", "n_v_max".
void read_faces() {
  int f,faces_start;
  // check whether section "_faces" was appeared earlier in the file (that is prohibited -- this section should appear only once)
  if (faces_done) rp_error("section '_faces' appears twice");
  // read the number of faces
  Nfaces=read_integer_from_file();
  // check whether this number is correct,
  if (Nfaces<4 || Nfaces>Nmax) rp_error("number of faces is out of range");
  // create the array "faces"
  faces=(struct face*) malloc((Nfaces+1)*sizeof(struct face));
  if (faces==NULL) rp_error("can't allocate memory");
  // create the array "n_v" of (Nfaces*10) elements
		n_v_Max=Nfaces*10;
		n_v=(int*) malloc(n_v_Max*sizeof(int));
  if (n_v==NULL) rp_error("can't allocate memory");
		n_v_max=0;
  // tell the program that no one of faces is read yet
  for (f=1; f<=Nfaces; f++) faces[f].read_from_file=0;
  // a loop: read words, one by one
  f=0; faces_start=0;
  while (1) {
    read_next_word();
    // if the word begins with '_', then section "_faces" is over (this word is a keyword that starts the next section)
    if (first_symbol=='_') break;
    // if the word begins with '@', then it starts information about a new face
    if (first_symbol=='@') {f=start_new_face(); faces_start=1; continue;}
    // otherwise, the word is a number of a vertex belonging to the current face
    if (faces_start) read_vertex_for_face(f);
  }
  // after reading the whole section "_faces", let's check that all faces were appeared in this section
		for (f=1; f<=Nfaces; f++)
			if (faces[f].read_from_file==0)
    rp_error1("face %d not found",f);
  // in conclusion, we let the program know that the section "_faces" is already read from file,
  faces_done=1;
  // and that the first word of the next section is read from file
  next_word_already_read=1;
}

// This function is called when a word is read (in section "_faces") that begins with symbol '@'.
// It means that information about a new face starts here, and the number of this face is written after '@'.
// The function picks this number of face, and lets the program know that reading of a new face is started.
int start_new_face() {
  int f;
  // pick the number of the face (variable "f")
  f=read_integer_after_at();
  // check whether this number is correct,
		if (f<1 || f>Nfaces) rp_error1("wrong number of face '%d'",f);
  // and whether this face was already appear in the file (that is prohibited -- each face should appear only once)
		if (faces[f].read_from_file!=0) rp_error1("face %d appears twice",f);
  // make necessary preparations for reading of vertices belonging to the new face
		faces[f].n_v_first=n_v_max;
		faces[f].nvertices=0;
  // tell the program that this face is already appears in the file
		faces[f].read_from_file=1;
  // return the number of the face
  return f;
}

// This function is called when a word is read (in section "_faces") that should contain a number of a vertex belonging to the face f.
// The function picks the number of the vertex and put it in appropriate place in the array "n_v".
void read_vertex_for_face(int f) {
  int v;
  // pick the number of the vertex (variable "v")
  next_word_already_read=1;
  v=read_integer_from_file();
  // check whether it is positive,
		if (v<1) rp_error2("wrong number of vertice '%d' in face %d",v,f);
  // and whether there is enough room in array "n_v" to store it
		if (n_v_max>=n_v_Max) rp_error("too many vertices (more than 10 vertices per face)");
  // then store it
		n_v[n_v_max]=v;
		n_v_max++;
		faces[f].nvertices++;
}


void read_next_word() {
  int mode=0;
  if (next_word_already_read) {next_word_already_read=0; return;}
  while (1) {
    read_next_word_1();
    if (word_match("_[")) {mode=1; continue;}
    if (word_match("_]")) {mode=0; continue;}
    if (mode==0) break;
  }
  first_symbol=s[0];
}

void read_next_word_1() {
  int n;
  n=fscanf(in,format_string,s);
  if (n<1 || s[0]==0) rp_error("keyword '_end' not found");
}

double read_double_from_file() {
  double dd; int n;
  read_next_word();
  n=sscanf(s,"%lf",&dd);
  if (n<1) rp_error1("word '%s' found where a number was expected",s);
  return dd;
}

int read_integer_from_file() {
  int ii, n;
  read_next_word();
  n=sscanf(s,"%d",&ii);
  if (n<1) rp_error1("word '%s' found where an integer number was expected",s);
  return ii;
}

int read_integer_after_at() {
  int ii, n;
  n=sscanf(s+1,"%d",&ii);
  if (n<1) rp_error1("word '%s' found where '@<integer_number>' was expected",s);
  return ii;
}

#undef rp_error
#undef rp_error1
#undef rp_error2
#undef word_match




